class Error(object):
    def __init__(self, message, code=None, location=None, field=None, **kwargs):
        self.message = message
        self.code = code
        self.location = location
        self.field = field
        self.__dict__.update(kwargs)

    def as_dict(self):
        data = self.__dict__.copy()

        for key in set(data.keys()):
            if data[key] is None:
                data.pop(key)

        return data


class APIError(Exception):
    status_code = 500
    error = Error('A server error occurred.', 'server_error')

    def __init__(self, error=None):
        if isinstance(error, str):
            self.error = Error(error)
        elif error:
            self.error = error


class BadRequest(APIError):
    status_code = 400
    error = Error('Invalid request.')


class AuthenticationFailed(APIError):
    status_code = 401
    error = Error('Incorrect authentication credentials.')


class NotAuthenticated(APIError):
    status_code = 401
    error = Error('Authentication credentials were not provided.')


class PermissionDenied(APIError):
    status_code = 403
    error = Error('You do not have permission to perform this action.')


class NotFound(APIError):
    status_code = 404
    error = Error('Not found.')


class NotAcceptable(APIError):
    status_code = 406
    error = Error('Could not satisfy the request Accept header.')


class UnsupportedMediaType(APIError):
    status_code = 415
    error = Error('Unsupported media type in request.')

    def __init__(self, media_type, error=None):
        if error:
            self.error = error
        else:
            self.error = Error(self.error.message)

        self.error.media_type = media_type



# Make marshmallow's functions and classes importable from flask-io
from marshmallow import pre_load, pre_dump, post_load, post_dump, Schema, ValidationError, validates_schema
from marshmallow.utils import missing

from .io import FlaskIO
from .errors import Error



"""
Custom Field classes that extend marshmallow Field class.

All fields from marshmallow has been imported here to allow the user import them from the flask-io.
"""


from marshmallow import fields
from marshmallow.fields import *
from .validate import Complexity, Length


__all__ = [
    'DelimitedList',
    'Enum',
    'Password',
    'String'
]


for field_name in (field_name for field_name in fields.__all__ if field_name not in __all__):
    __all__.append(field_name)


class DelimitedList(List):
    """
    A delimited list composed with another `Field` class that loads from a delimited string.
    """

    delimiter = ','

    def __init__(self, cls_or_instance, delimiter=None, **kwargs):
        """
        Initializes a new instance of `DelimitedList`.

        :param Field cls_or_instance: A field class or instance.
        :param str delimiter: Delimiter between values.
        """
        self.delimiter = delimiter or self.delimiter
        super().__init__(cls_or_instance, **kwargs)

    def _serialize(self, value, attr, obj):
        values = super()._serialize(value, attr, obj)
        return self.delimiter.join(format(v) for v in values)

    def _deserialize(self, value, attr, data):
        values = value.split(self.delimiter)
        return super()._deserialize(values, attr, data)


class Enum(Field):
    """
    A field that provides a set of enumerated values which an attribute must be constrained to.
    """

    def __init__(self, enum_type, *args, **kwargs):
        """
        Initializes a new instance of `Enum`.

        :param enum_type: A Python enum class.
        """

        super().__init__(*args, **kwargs)
        self.enum_type = enum_type
        self.__member_type = type(list(self.enum_type)[0].value)

    def _serialize(self, value, attr, obj):
        try:
            if type(value) is self.enum_type:
                return value.value
            if type(value) is not self.__member_type:
                value = self.__member_type(value)
            return self.enum_type(value).value
        except:
            self.fail('validator_failed')

    def _deserialize(self, value, attr, data):
        try:
            if type(value) is self.enum_type:
                return value
            if type(value) is not self.__member_type:
                value = self.__member_type(value)
            return self.enum_type(value)
        except:
            self.fail('validator_failed')


class Password(Field):
    """
    A password field used to validate strong passwords.
    """

    def __init__(self, upper=1, lower=1, letters=1, digits=1, special=1, special_chars=None, min_length=6, max_length=None, *args, **kwargs):
        """
        Initializes a new instance of `Password`.

        :param int upper: Number of uppercase letters that the password must have.
        :param int lower: Number of lowercase letters that the password must have.
        :param int letters: Number of letters that the password must have.
        :param int letters: Number of special characters that the password must have.
        :param str special_chars: List of special characters allowed.
        :param int min_length: Minimum of characters that the password must have.
        :param int max_length: Maximum of characters that the password must have.
        """
        super().__init__(*args, **kwargs)

        self.validators.append(Length(min_length, max_length))
        self.validators.append(Complexity(upper, lower, letters, digits, special, special_chars))


class String(fields.String):
    """
    Extends Marshmallow String Field to add new features.
    """

    allow_empty = True
    none_if_empty = False
    strip = False
    only_numeric = False

    default_error_messages = {
        'empty': 'Field may not be empty.',
        'only_numeric': 'Only numeric chars are allowed.'
    }

    def __init__(self, allow_empty=None, none_if_empty=None, strip=None, only_numeric=None, *args, **kwargs):
        """
        Initializes a new instance of 'String'.

        :param bool allow_empty: Indicates whether the string can be empty ('').
        :param bool none_if_empty: Indicates whether the empty ('') should be considered 'None'.
        :param bool strip: Indicates whether the string should be trimmed.
        :param bool only_numeric: Indicates whether the string should have only numbers
        :return:
        """

        super().__init__(*args, **kwargs)
        self.allow_empty = self.allow_empty if allow_empty is None else allow_empty
        self.strip = self.strip if strip is None else strip
        self.none_if_empty = self.none_if_empty if none_if_empty is None else none_if_empty
        self.only_numeric = self.only_numeric if only_numeric is None else only_numeric

    def deserialize(self, value, attr=None, data=None):
        if self.strip and value:
            value = value.strip()

        if self.none_if_empty and value == '':
            value = None

        return super().deserialize(value, attr, data)

    def _validate(self, value):
        if not self.allow_empty and value == '':
            self.fail('empty')

        if self.none_if_empty and not self.allow_none and value is None:
            self.fail('null')

        if self.only_numeric and value and not value.isnumeric():
            self.fail('only_numeric')

        super()._validate(value)


class UUID(fields.UUID):
    """
    Extends Marshmallow UUID Field to add new features.
    """

    def __init__(self, as_text=False, *args, **kwargs):
        """
        Initializes a new instance of 'String'.

        :param bool as_text: Indicates whether the value should be converted to string.
        """
        super().__init__(*args, **kwargs)
        self.as_text = as_text

    def _deserialize(self, value, attr, data):
        value_as_uuid = super()._deserialize(value, attr, data)

        if self.as_text:
            return value

        return value_as_uuid


# Aliases
Str = String



"""
Parsers used to parse a byte array into Python object.
"""

try:
    import simplejson as json
except ImportError:
    import json

from abc import ABCMeta, abstractmethod
from .mimetypes import MimeType


class Parser(metaclass=ABCMeta):
    """
    Base class for all parsers.
    """

    mimetype = None

    @abstractmethod
    def parse(self, data, mimetype):
        """
        Parses the given data and returns a Python object.
        :param data: The data to be parsed.
        :param mimetype: The mimetype to parse the data.
        :return: A Python object
        """
        pass


class JSONParser(Parser):
    """
    Parses JSON data into Python object.
    """

    mimetype = MimeType('application/json')

    def parse(self, data, mimetype):
        """
        Parses a byte array containing a JSON document and returns a Python object.
        :param data: The byte array containing a JSON document.
        :param MimeType mimetype: The mimetype chose to parse the data.
        :return: A Python object.
        """
        encoding = mimetype.params.get('charset') or 'utf-8'

        return json.loads(data.decode(encoding))



"""
Handling of mimetypes from the HTTP request.
"""


class MimeType(object):
    """
    Represents a mimetype.
    """

    def __init__(self, mimetype):
        """
        Initializes a new instance of MimeType.

        :param str mimetype: The mimetype.
        """
        self.mimetype = mimetype
        self.full_type, self.params = self.__parse_mimetype(mimetype)
        self.main_type, sep, self.sub_type = self.full_type.partition('/')

    def match(self, other):
        """
        Checks if the given MimeType matches to the this MimeType.

        :param MimeType other: The MimeType to compare to.
        :return bool: True if this MimeType matches the given MediaType.
        """
        if self.sub_type != '*' and other.sub_type != '*' and other.sub_type != self.sub_type:
            return False

        if self.main_type != '*' and other.main_type != '*' and other.main_type != self.main_type:
            return False

        return True

    def __parse_mimetype(self, mimetype):
        """
        Extracts the full type and parameters from the given MimeType.
        :param str mimetype: The mimetype to be parsed.
        :return: Returns a tuple with full type and parameters.
        """
        plist = mimetype.split(';')

        full_type = plist.pop(0).lower().strip()
        params = {}

        for p in plist:
            kv = p.split('=')
            if len(kv) != 2:
                continue
            v = kv[1].strip()
            if v:
                params[kv[0].strip()] = v

        return full_type, params



"""

"""

from flask import request
from flask_io import errors


class Action(object):
    def __init__(self, func, default_authenticators, default_permissions, trace_enabled):
        self.func = func

        self.authenticators = default_authenticators
        if hasattr(func, 'authenticators'):
            self.authenticators = func.authenticators

        self.permissions = default_permissions
        if hasattr(func, 'permissions'):
            self.permissions = func.permissions

        self.trace_enabled = trace_enabled

    def __call__(self, *args, **kwargs):
        self.perform_authentication()
        self.perform_authorization()

        return self.func(*args, **kwargs)

    def perform_authentication(self):
        """
        Perform authentication on the incoming request.
        """

        if not self.authenticators:
            return

        request.user = None
        request.auth = None

        for authenticator in self.authenticators:
            auth_tuple = authenticator.authenticate()

            if auth_tuple:
                request.user = auth_tuple[0]
                request.auth = auth_tuple[1]
                break

    def perform_authorization(self):
        """
        Check if the request should be permitted.
        Raises an appropriate exception if the request is not permitted.
        """

        for permission in self.permissions:
            if not permission.has_permission():
                if request.user:
                    raise errors.PermissionDenied()
                else:
                    raise errors.NotAuthenticated()



import functools

from flask import request
from inspect import isclass
from logging import getLogger
from werkzeug.exceptions import HTTPException
from . import fields, missing, ValidationError
from .actions import Action
from .errors import APIError, BadRequest, NotAcceptable, UnsupportedMediaType
from .negotiation import DefaultContentNegotiation
from .parsers import JSONParser
from .renderers import JSONRenderer
from .tracing import Tracer
from .utils import errors_to_dict, get_fields_from_request, http_status_message, marshal, unpack, \
    validation_error_to_errors, Stopwatch


class FlaskIO(object):
    """
    The class responsible for parsing request into function parameters and deserialize function returns into response.
    """

    def __init__(self, app=None):
        """
        Initializes a new instance.

        :param app: A Flask instance class.
        """

        self.__app = None

        self.content_negotiation = DefaultContentNegotiation()
        self.default_authenticators = []
        self.default_permissions = []
        self.default_parsers = [JSONParser()]
        self.default_renderers = [JSONRenderer()]

        self.logger = getLogger('flask-io')

        self.tracer = Tracer(self)

        if app:
            self.init_app(app)

    def init_app(self, app):
        """
        Initialize this class with the specified :class:`flask.Flask` application

        :param app: The Flask application.
        """

        self.__app = app
        self.__app.before_first_request(self.__setup)

        self.tracer.enabled = self.__app.config.get('TRACE_ENABLED', self.tracer.enabled)

    def bad_request(self, error):
        """
        Gets a 404 response with the specified error.

        :param error: The error to include in the response.
        :return: A Flask response object.
        """

        return self.__make_response((errors_to_dict(error), 400))

    def conflict(self, error):
        """
        Gets a 409 response with the specified error.

        :param error: The error to include in the response.
        :return: A Flask response object.
        """

        return self.__make_response((errors_to_dict(error), 409))

    def created(self, data, schema=None, envelope=None):
        """
        Gets a 201 response with the specified data.

        :param data: The content value.
        :param schema: The schema to serialize the data.
        :param envelope: The key used to envelope the data.
        :return: A Flask response object.
        """

        data = marshal(data, schema, envelope)
        return self.__make_response((data, 201))

    def forbidden(self, error):
        """
        Gets a 403 response with the specified error.

        :param error: The error to include in the response.
        :return: A Flask response object.
        """

        return self.__make_response((errors_to_dict(error), 403))

    def no_content(self):
        """
        Gets a 204 response with no content.

        :return: A Flask response object.
        """

        return self.__make_response((None, 204))

    def not_found(self, error):
        """
        Gets a 404 response with the specified error.

        :param error: The error to include in the response.
        :return: A Flask response object.
        """

        return self.__make_response((errors_to_dict(error), 404))

    def ok(self, data, schema=None, envelope=None):
        """
        Gets a 200 response with the specified data.

        :param data: The content value.
        :param schema: The schema to serialize the data.
        :param envelope: The key used to envelope the data.
        :return: A Flask response object.
        """

        data = marshal(data, schema, envelope)
        return self.__make_response(data)

    def unauthorized(self, error):
        """
        Gets a 401 response with the specified error.

        :param error: The error to include in the response.
        :return: A Flask response object.
        """
        return self.__make_response((errors_to_dict(error), 401))

    def authenticators(self, auths):
        """
        A decorator that sets a list of authenticators for a function.

        :param auths: The list of authenticator instances or classes.
        :return: A function
        """

        if not isinstance(auths, (list, tuple)):
            auths = [auths]

        instances = []

        for auth in auths:
            if isclass(auth):
                instances.append(auth())
            else:
                instances.append(auth)

        def decorator(func):
            func.authenticators = instances
            return func
        return decorator

    def permissions(self, perms):
        """
        A decorator that sets a list of permissions for a function.

        :param perms: The list of permission instances or classes.
        :return: A function
        """

        if not isinstance(perms, (list, tuple)):
            perms = [perms]

        instances = []

        for perm in perms:
            if isclass(perm):
                instances.append(perm())
            else:
                instances.append(perm)

        def decorator(func):
            func.permissions = perms
            return func
        return decorator

    def from_body(self, param_name, schema):
        """
        A decorator that converts the request body into a function parameter based on the specified schema.

        :param param_name: The parameter which receives the argument.
        :param schema: The schema class or instance used to deserialize the request body toa Python object.
        :return: A function
        """

        schema = schema() if isclass(schema) else schema

        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                kwargs[param_name] = self.__parse_body(schema)
                return func(*args, **kwargs)
            return wrapper
        return decorator

    def from_cookie(self, param_name, field):
        """
        A decorator that converts a request cookie into a function parameter based on the specified field.

        :param str param_name: The parameter which receives the argument.
        :param Field field: The field class or instance used to deserialize the request cookie to a Python object.
        :return: A function
        """

        return self.__from_source(param_name, field, lambda: request.cookies, 'cookie')

    def from_form(self, param_name, field):
        """
        A decorator that converts a request form into a function parameter based on the specified field.

        :param str param_name: The parameter which receives the argument.
        :param Field field: The field class or instance used to deserialize the request form to a Python object.
        :return: A function
        """
        return self.__from_source(param_name, field, lambda: request.form, 'form')

    def from_header(self, param_name, field):
        """
        A decorator that converts a request header into a function parameter based on the specified field.

        :param str param_name: The parameter which receives the argument.
        :param Field field: The field class or instance used to deserialize the request header to a Python object.
        :return: A function
        """
        return self.__from_source(param_name, field, lambda: request.headers, 'header')

    def from_query(self, param_name, field):
        """
        A decorator that converts a query string into a function parameter based on the specified field.

        :param param_name: The parameter which receives the argument.
        :param Field field: The field class or instance used to deserialize the request query string to a Python object.
        :return: A function
        """

        return self.__from_source(param_name, field, lambda: request.args, 'query')

    def marshal_with(self, schema, envelope=None):
        """
        A decorator that apply marshalling to the return values of your methods.

        :param schema: The schema class to be used to serialize the values.
        :param envelope: The key used to envelope the data.
        :return: A function.
        """

        # schema is pre instantiated to avoid instantiate it
        # on every request
        schema_is_class = isclass(schema)
        schema_cache = schema() if schema_is_class else schema

        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                data = func(*args, **kwargs)
                if isinstance(data, self.__app.response_class):
                    return data

                schema_instance = schema_cache

                # if there is the parameter 'fields' in the url
                # we cannot use the cached schema instance
                # in this case we have to instantiate the schema
                # on every request.
                if schema_is_class:
                    only = get_fields_from_request()
                    if only:
                        schema_instance = schema(only=only)

                return marshal(data, schema_instance, envelope)
            return wrapper
        return decorator

    def trace_inspect(self):
        """
        A decorator that allows to inspect/change the trace data.
        """
        def decorator(f):
            self.tracer.inspector = f
            return f
        return decorator

    def trace_emit(self):
        """
        A decorator that allows to change the trace emitter.
        By default Python logging is used to emit the trace data.
        """
        def decorator(f):
            self.tracer.emitter = f
            return f
        return decorator

    def __handle_error(self, e):
        if isinstance(e, ValidationError):
            code = 400
            error = validation_error_to_errors(e)
        elif isinstance(e, APIError):
            code = e.status_code
            error = e.error
        elif isinstance(e, HTTPException):
            code = e.code
            error = getattr(e, 'description', http_status_message(code))
        else:
            code = 500
            error = str(e) if self.__app.config.get('DEBUG') else http_status_message(code)
            self.logger.error(str(e))

        errors_data = errors_to_dict(error)

        return self.__make_response((errors_data, code), self.default_renderers[0])

    def __make_response(self, data, default_renderer=None):
        """
        Creates a Flask response object from the specified data.
        The appropriated encoder is taken based on the request header Accept.
        If there is not data to be serialized the response status code is 204.

        :param data: The Python object to be serialized.
        :return: A Flask response object.
        """

        status = headers = None
        if isinstance(data, tuple):
            data, status, headers = unpack(data)

        if data is None:
            data = self.__app.response_class(status=204)
        elif not isinstance(data, self.__app.response_class):
            renderer, mimetype = self.content_negotiation.select_renderer(request, self.default_renderers)

            if not renderer:
                if not default_renderer:
                    raise NotAcceptable()

                renderer = default_renderer
                mimetype = default_renderer.mimetype

            data_bytes = renderer.render(data, mimetype)
            data = self.__app.response_class(data_bytes, mimetype=mimetype.mimetype)

        if status is not None:
            data.status_code = status

        if headers:
            data.headers.extend(headers)

        return data

    def __from_source(self, param_name, field, getter_data, location):
        field = field() if isclass(field) else field
        if not field.required:
            field.allow_none = True

        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                kwargs[param_name] = self.__parse_field(param_name, field, getter_data(), location)
                return func(*args, **kwargs)
            return wrapper
        return decorator

    def __parse_field(self, field_name, field, data, location):
        field.allow_none = True

        field_name = field.load_from or field_name

        if isinstance(field, fields.List):
            raw_value = data.getlist(field_name) or missing
        else:
            raw_value = data.get(field_name) or missing

        if raw_value is missing:
            missing_value = field.missing
            raw_value = missing_value() if callable(missing_value) else missing_value

        if raw_value is missing and not field.required:
            raw_value = None

        try:
            return field.deserialize(raw_value, field_name, data)
        except ValidationError as e:
            e.messages = {field_name: e.messages}
            e.kwargs['location'] = location
            raise

    def __parse_body(self, schema):
        if not request.data:
            raise BadRequest('Payload missing.')

        parser, mimetype = self.content_negotiation.select_parser(request, self.default_parsers)

        if not parser:
            raise UnsupportedMediaType(request.headers['content-type'])

        try:
            decoded_data = parser.parse(request.data, mimetype)
        except:
            raise BadRequest('Malformed request.')

        model, errors = schema.load(decoded_data)

        if errors:
            raise ValidationError(errors, data=request.data, location='body')

        return model

    def __process_action(self, action):
        def decorator(**kwargs):
            latency = response = error = None

            if action.trace_enabled and self.tracer.enabled:
                latency = Stopwatch.start_new()

            try:
                response = action(**kwargs)
                response = self.__make_response(response)
                return response
            except Exception as e:
                error = e
                response = self.__handle_error(e)
                return response
            finally:
                if action.trace_enabled and self.tracer.enabled:
                    latency.stop()
                    self.tracer.trace(request, response, error, latency)

        return decorator

    def __setup(self):
        for endpoint in self.__app.view_functions.keys():
            for rule in self.__app.url_map.iter_rules(endpoint):
                if self.tracer.match(rule):
                    trace_enabled = True
                    break
            else:
                trace_enabled = False

            action = Action(self.__app.view_functions[endpoint],
                            self.default_authenticators,
                            self.default_permissions,
                            trace_enabled)

            self.__app.view_functions[endpoint] = self.__process_action(action)



"""
Provides various authentication policies.
"""

from abc import ABCMeta, abstractmethod


class Authenticator(metaclass=ABCMeta):
    """
    A base class from which all authenticator classes should inherit.
    """

    @abstractmethod
    def authenticate(self):
        """
        Authenticate the request and return a two-tuple of (user, token).
        """
        pass



"""
Content negotiation selects a appropriated parser and renderer for a HTTP request.
"""

from abc import ABCMeta, abstractmethod
from .mimetypes import MimeType


class ContentNegotiation(metaclass=ABCMeta):
    """
    Base class for all content negotiations.
    """

    @abstractmethod
    def select_parser(self, request, parsers):
        """
        Selects the appropriated parser for the given request.
        :param request: The HTTP request.
        :param parsers: The lists of parsers.
        :return: The parser selected or none.
        """
        pass

    @abstractmethod
    def select_renderer(self, request, renderers):
        """
        Selects the appropriated renderer for the given request.
        :param request: The HTTP request.
        :param renderers: The lists of renderers.
        :return: The renderer selected or none.
        """
        pass


class DefaultContentNegotiation(ContentNegotiation):
    """
    Selects a parser by request content type and a
    renderer by request accept.
    """

    def select_parser(self, request, parsers):
        """
        Selects the appropriated parser which matches to the request's content type.
        :param request: The HTTP request.
        :param parsers: The lists of parsers.
        :return: The parser selected or none.
        """

        if not request.content_type:
            return parsers[0], parsers[0].mimetype

        mimetype = MimeType(request.content_type)

        for parser in parsers:
            if mimetype.match(parser.mimetype):
                return parser, mimetype

        return None, None

    def select_renderer(self, request, renderers):
        """
        Selects the appropriated parser which matches to the request's accept.
        :param request: The HTTP request.
        :param renderers: The lists of parsers.
        :return: The parser selected or none.
        """

        if not len(request.accept_mimetypes):
            return renderers[0], renderers[0].mimetype

        for mimetype, quality in request.accept_mimetypes:
            accept_mimetype = MimeType(mimetype)
            for renderer in renderers:
                if accept_mimetype.match(renderer.mimetype):
                    return renderer, accept_mimetype

        return None, None



"""
Renderers used to render a Python object into byte array.
"""

from abc import ABCMeta, abstractmethod
from . import json
from .mimetypes import MimeType


class Renderer(metaclass=ABCMeta):
    """
    Base class for all renderers.
    """

    mimetype = None

    @abstractmethod
    def render(self, data, mimetype):
        """
        Render the given data into JSON and returns a byte array.
        :param data: The data to be rendered.
        :param mimetype: The mimetype to render the data.
        :return: A byte array.
        """
        pass


class JSONRenderer(Renderer):
    """
    Renderer which render into JSON.
    """

    mimetype = MimeType('application/json')

    def render(self, data, mimetype):
        """
        Serializes a Python object into a byte array containing a JSON document.
        :param data: A Python object.
        :param mimetype: The mimetype to render the data.
        :return: A byte array containing a JSON document.
        """

        indent = self.__get_indent(mimetype)
        encoding = mimetype.params.get('charset') or 'utf-8'
        return json.dumps(data, indent=indent).encode(encoding)

    def __get_indent(self, mimetype):
        """
        Gets the indent parameter from the mimetype.
        :param MimeType mimetype: The mimetype with parameters.
        :return int: The indent if found, otherwise none.
        """
        indent = max(int(mimetype.params.get('indent', '0')), 0)

        if indent == 0:
            return None

        return indent



"""
Provides a set of pluggable permission policies.
"""

from abc import ABCMeta, abstractmethod
from flask import request


class Permission(metaclass=ABCMeta):
    """
    A base class from which all permission classes should inherit.
    """

    @abstractmethod
    def has_permission(self):
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        pass


class IsAuthenticated(Permission):
    """
    Allows access only to authenticated users.
    """

    def has_permission(self):
        if not hasattr(request, 'user'):
            return False
        return request.user is not None
